@@ -28,6 +28,7 @@
 import re
 
 from pyparsing import (
+    AtLineStart,
     Combine,
     Forward,
     Group,
@@ -57,6 +58,9 @@ def escape_to_python(string):
     pystring = re.sub(
         "\\\\n", "\n", pystring
     )  # Convert escaped newline to a real newline
+    pystring = re.sub(
+        "\\\\r", "\r", pystring
+    )  # Convert escaped carriage-return to a real carriage-return
     pystring = re.sub("\\\\t", "\t", pystring)  # Convert escape tab to a real tab
     pystring = re.sub(
         "\\\\\\\\", "\\\\", pystring
@@ -74,11 +78,24 @@ def extract_text(values):
     return "".join(result)
 
 
+def extract_id(values):
+    for value in values:
+        if isinstance(value, str) and value.startswith('"'):
+            continue
+        else:
+            if isinstance(value, str):
+                return value
+            break
+
+    return "UNKNOWN_ID"
+
+
 def escape_to_rc(string):
     """Escape a given Python string into a valid .rc string."""
     rcstring = re.sub("\\\\", "\\\\\\\\", string)
     rcstring = re.sub("\t", "\\\\t", rcstring)
     rcstring = re.sub("\n", "\\\\n", rcstring)
+    rcstring = re.sub("\r", "\\\\r", rcstring)
     return rcstring
 
 
@@ -153,7 +170,7 @@ def rc_statement():
 
     comments = c_style_comment ^ one_line_comment
 
-    precompiler = Word("#", alphanums) + rest_of_line
+    precompiler = AtLineStart(Word("#", alphanums) + rest_of_line)
 
     language_definition = (
         "LANGUAGE"
@@ -164,19 +181,19 @@ def rc_statement():
     block_start = (Keyword("{") | Keyword("BEGIN")).set_name("block_start")
     block_end = (Keyword("}") | Keyword("END")).set_name("block_end")
 
-    name_id = Group(Word(alphas, alphanums + "_")).set_name("name_id")
-
     numbers = Word(nums)
 
     integerconstant = numbers ^ Combine("0x" + numbers)
 
+    name_id = Group(Word(alphas, alphanums + "_") | integerconstant)
+
     constant = Combine(
-        Optional(Keyword("NOT")) + (name_id | integerconstant),
+        Optional(Keyword("NOT")) + Group(name_id),
         adjacent=False,
         join_string=" ",
     )
 
-    combined_constants = delimited_list(constant, "|")
+    combined_constants = delimited_list(constant, "|", min=2)
 
     concatenated_string = OneOrMore(quoted_string)
 
@@ -189,6 +206,7 @@ def rc_statement():
     undefined_control = (
         Group(
             name_id.set_results_name("id_control")
+            + Optional(",")
             + delimited_list(
                 concatenated_string ^ constant ^ numbers ^ Group(combined_constants)
             ).set_results_name("values_")
@@ -221,7 +239,9 @@ def rc_statement():
         Keyword("POPUP")("block_type")
         + Optional(quoted_string("caption"))
         + block_start
-        + ZeroOrMore(Group(menu_item | popup_block), stop_on=block_end)("elements")
+        + ZeroOrMore(Group(menu_item | popup_block | comments), stop_on=block_end)(
+            "elements"
+        )
         + block_end
     )("popups*")
 
@@ -230,7 +250,7 @@ def rc_statement():
         + Keyword("MENU")("block_type")
         + block_options
         + block_start
-        + ZeroOrMore(popup_block, stop_on=block_end)
+        + ZeroOrMore(popup_block | comments, stop_on=block_end)
         + block_end
     )
 
@@ -239,7 +259,7 @@ def rc_statement():
 
 def generate_stringtable_name(identifier):
     """Return the name generated for a stringtable element."""
-    return "STRINGTABLE." + identifier
+    return f"STRINGTABLE.{identifier}"
 
 
 def generate_menu_pre_name(block_type, block_id):
@@ -261,7 +281,7 @@ def generate_popup_pre_name(pre_name, caption):
 
 def generate_popup_caption_name(pre_name):
     """Return the name generated for a caption of a popup."""
-    return "%s.POPUP.CAPTION" % (pre_name)
+    return f"{pre_name}.POPUP.CAPTION"
 
 
 def generate_menuitem_name(pre_name, block_type, identifier):
@@ -271,7 +291,7 @@ def generate_menuitem_name(pre_name, block_type, identifier):
 
 def generate_dialog_caption_name(block_type, identifier):
     """Return the name generated for a caption of a dialog."""
-    return "{}.{}.{}".format(block_type, identifier, "CAPTION")
+    return f"{block_type}.{identifier}.CAPTION"
 
 
 def generate_dialog_control_name(block_type, block_id, control_type, identifier):
@@ -295,16 +315,21 @@ class rcfile(base.TranslationStore):
     UnitClass = rcunit
     default_encoding = "cp1252"
 
-    def __init__(self, inputfile=None, lang=None, sublang=None, **kwargs):
+    def __init__(
+        self, inputfile=None, lang=None, sublang=None, encoding=None, **kwargs
+    ):
         """Construct an rcfile, optionally reading in from inputfile."""
-        super().__init__(**kwargs)
+        super().__init__(encoding=encoding, **kwargs)
         self.filename = getattr(inputfile, "name", "")
         self.lang = lang
         self.sublang = sublang
         if inputfile is not None:
             rcsrc = inputfile.read()
             inputfile.close()
-            self.parse(rcsrc)
+            self.parse(
+                rcsrc,
+                encoding=encoding or "auto",
+            )
 
     def add_popup_units(self, pre_name, popup):
         """Transverses the popup tree making new units as needed."""
@@ -324,7 +349,7 @@ class rcfile(base.TranslationStore):
                     if newtext:
                         newunit = rcunit(newtext)
                         newunit.name = generate_menuitem_name(
-                            pre_name, element.block_type, element.values_[1]
+                            pre_name, element.block_type, extract_id(element.values_)
                         )
                         newunit.match = element
                         self.addunit(newunit)
@@ -361,6 +386,7 @@ class rcfile(base.TranslationStore):
             if statement[0] == "#pragma" and "code_page" in statement[1]:
                 expected_encoding = parse_encoding_pragma(statement[1])
                 if expected_encoding and expected_encoding != self.encoding:
+                    self.units = []
                     self.parse(rcsrc, expected_encoding)
                     return
             if statement.language:
@@ -418,7 +444,7 @@ class rcfile(base.TranslationStore):
                                     statement.block_type,
                                     statement.block_id[0],
                                     control.id_control[0],
-                                    control.values_[1],
+                                    extract_id(control.values_),
                                 )
                                 newunit.match = control
                                 self.addunit(newunit)
